"""Strangeworks Braket SDK"""
import importlib.metadata

from .device import StrangeworksDevice  # noqa: F401
from .job import StrangeworksQuantumJob  # noqa: F401
from .task import StrangeworksQuantumTask  # noqa: F401


__version__ = importlib.metadata.version("strangeworks-braket")



from __future__ import annotations

import asyncio
import json
import os
import time
from typing import Any, Dict, Optional

import strangeworks as sw
from braket.tasks.quantum_task import QuantumTask
from strangeworks_core.errors.error import StrangeworksError
from strangeworks_core.types.job import Job, Status


class StrangeworksQuantumJob(QuantumTask):
    _product_slug = "amazon-braket"

    def __init__(self, job: Job, *args, **kwargs):
        self.job: Job = job

    @property
    def id(self) -> str:
        """The id of the job.

        Returns
        -------
        id: str
            The id of the job. This is the id of the job in Strangeworks.
        """
        return self.job.slug

    def cancel(self) -> None:
        """Cancel the job.

        Raises
        ------
        StrangeworksError
            If the job has not been submitted yet.

        """
        if not self.job.external_identifier:
            raise StrangeworksError(
                "Job has not been submitted yet. Missing external_identifier."  # noqa: E501
            )

        resource = sw.get_resource_for_product(StrangeworksQuantumJob._product_slug)
        cancel_url = (
            f"{resource.proxy_url()}/hybrid_jobs/{self.job.external_identifier}"
        )
        # todo: strangeworks-python is rest_client an optional thing. i dont think it should be # noqa: E501
        # this is something we should discuss
        sw.client.rest_client.delete(url=cancel_url)

    def state(self) -> str:
        """Get the state of the job.

        Returns
        -------
        state: str
            The state of the job.

        Raises
        ------
        StrangeworksError
            If the job has not been submitted yet.
            Or if we find are not able to find the status.
        """
        if not self.job.external_identifier:
            raise StrangeworksError(
                "Job has not been submitted yet. Missing external_identifier."  # noqa: E501
            )

        res = sw.execute_get(
            StrangeworksQuantumJob._product_slug,
            f"hybrid_jobs/{self.job.external_identifier}",
        )
        self.job = StrangeworksQuantumJob._transform_dict_to_job(res)

        if not self.job.remote_status:
            raise StrangeworksError("Job has no status")
        return self.job.remote_status

    def result(self) -> Dict[str, Any]:
        """Get the result of the job.

        Returns
        -------
        result: BraketSchemaBase
            The result of the job.

        Raises
        ------
        StrangeworksError
            If the job has not been submitted yet.
            Or if the job did not complete successfully.
            Or unable to fetch the results for the job.
        """
        if not self.job.external_identifier:
            raise StrangeworksError(
                "Job has not been submitted yet. Missing external_identifier."  # noqa: E501
            )
        while self.job.status not in {
            Status.COMPLETED,
            Status.FAILED,
            Status.CANCELLED,
        }:
            res = sw.execute_get(
                StrangeworksQuantumJob._product_slug,
                f"hybrid_jobs/{self.job.external_identifier}",
            )
            self.job = StrangeworksQuantumJob._transform_dict_to_job(res)
            time.sleep(2.5)

        if self.job.status != Status.COMPLETED:
            errors = sw.get_error_messages(job_slug=self.id)

            msg = f"Job {self.id} has status {self.job.status} \n"
            msg = "\nLogs from Braket: \n"
            msg += "\n************************************\n\n"
            for line in errors["parent_job"][0]:
                msg += str(line)
                msg += "\n"
            msg += "\n************************************\n\n"
            msg += "Job did not complete successfully. Any logs are shown above"
            raise StrangeworksError(msg)
        # sw.jobs will return type errors until it updates their type hints
        # todo: update strangeworks-python job type hints
        # todo: at this point in time, sw.jobs returns a different type than sw.execute
        jobs = sw.jobs(slug=self.job.slug)
        if not jobs:
            raise StrangeworksError("Job not found.")
        if len(jobs) != 1:
            raise StrangeworksError("Multiple jobs found.")
        job: Job = jobs[0]
        if not job.files:
            raise StrangeworksError("Job has no files.")
        # for now the strangeworks-python library still returns the Job.files as Files not JobFiles # noqa: E501
        files = list(
            filter(lambda f: f.file_name == "job_results_braket.json", job.files)
        )
        if len(files) != 1:
            raise StrangeworksError("Job has multiple files")

        file = files[0]
        if not file.url:
            raise StrangeworksError("Job file has no url")
        # why does this say it returns a list of files?
        # did it not just download the file?
        # is the contents not some dictionary?
        # todo probably have to update this in strangeworks-python
        contents = sw.download_job_files([file.url])
        if not contents:
            raise StrangeworksError("Unable to download result file.")
        if len(contents) != 1:
            raise StrangeworksError("Unable to download result file.")
        job_result = contents[0]

        return job_result

    def async_result(self) -> asyncio.Task:
        raise NotImplementedError

    def metadata(self, use_cached_value: bool = False) -> Dict[str, Any]:
        raise NotImplementedError

    @staticmethod
    def from_strangeworks_slug(id: str) -> StrangeworksQuantumJob:
        """Get a job from a strangeworks id.

        Parameters
        ----------
        id: str
            The strangeworks id of the job.

        Returns
        -------
        job: StrangeworksQuantumJob
            The job.

        Raises
        ------
        StrangeworksError
            If no job is found for the id.
            Or if multiple job are found for the id.
        """
        # todo: at this point in time, sw.jobs returns a different type than sw.execute
        jobs = sw.jobs(slug=id)
        if not jobs:
            raise StrangeworksError("No jobs found for slug")
        if len(jobs) != 1:
            raise StrangeworksError("Multiple jobs found for slug")
        job = jobs[0]
        return StrangeworksQuantumJob(job)

    @staticmethod
    def create_hybrid(
        device_arn: str,
        filepath: str,
        hyperparameters: Dict[str, Any],
        device_parameters: Optional[Dict[str, Any]] = None,
        tags: Optional[Dict[str, str]] = None,
        *args,
        **kwargs,
    ) -> StrangeworksQuantumJob:
        """Create a job.

        Parameters
        ----------
        device_arn: str
            The name of the device to run the job on.
        filepath: str
            Path to the python file that will be run.
        hyperparameters: Dict[str, Any]
            Dictionary of hyperparameters to pass to the job.
            Must be json serializable.
        device_parameters: Optional[Dict[str, Any]]
            The device parameters.
        tags: Optional[Dict[str, str]]
            The tags to add to the strangeworks job.

        Returns
        -------
        job: StrangeworksQuantumJob
            The job.

        Raises
        ------
        StrangeworksError
            If the job specification is not a circuit, or openqasm program.

        """
        file_list = []
        with open(filepath) as fh:
            for line in fh:
                file_list.append(line)
        json_file = json.dumps(file_list)

        payload = {
            "pythonfile": json_file,
            "hyperparameters": hyperparameters,
            "aws_device_arn": device_arn,
            "device_parameters": device_parameters if device_parameters else {},
        }

        res = sw.execute_post(
            StrangeworksQuantumJob._product_slug, payload, endpoint="hybrid_jobs"
        )
        sw_job = StrangeworksQuantumJob._transform_dict_to_job(res)
        if tags is not None:
            sw.add_tags(sw_job.slug, tags)

        return StrangeworksQuantumJob(sw_job)

    @staticmethod
    def create(
        source_module: str,
        hyperparameters: Dict[str, Any],
        device: str,
        job_name: str,
    ) -> StrangeworksQuantumJob:
        """Create a job.

        Parameters
        ----------
        source_module: str
            The directory or single file containing the code to run.
        entry_point: str
            The main script or function the job will run.
        hyperparameters: Dict[str, Any]
            Dictionary of hyperparameters to pass to the job.
            Must be json serializable.
        device: str
            The ARN of the device to run the job on.
        job_name: str
            The name of the job, which will be added as a tag.
        input_data: str
            The input data for the job.
        wait_until_complete: bool, optional
            Whether to wait until the job is complete. Defaults to False.

        Returns
        -------
        job: StrangeworksQuantumJob
            The job.

        Raises
        ------
        StrangeworksError
            If the job specification is not a circuit, or openqasm program.

        """
        # Check if the input is suitable for create_hybrid
        if not os.path.isfile(source_module):
            raise StrangeworksError("source_module must be a valid file path")

        return StrangeworksQuantumJob.create_hybrid(
            device_arn=device,
            filepath=source_module,
            hyperparameters=hyperparameters,
            tags=[job_name],
        )

    # create a method that transforms the dict into a job
    # first it must convert the json keys from snake_case to camelCase
    # then it must create a job from the dict
    @staticmethod
    def _transform_dict_to_job(d: Dict[str, Any]) -> Job:
        # todo: this is unfortunate. dont like that we need to do this.
        def to_camel_case(snake_str):
            components = snake_str.split("_")
            # We capitalize the first letter of each component except the first one
            # with the 'title' method and join them together.
            return components[0] + "".join(x.title() for x in components[1:])

        remix = {to_camel_case(key): value for key, value in d.items()}
        return Job.from_dict(remix)



from __future__ import annotations

import asyncio
import json
import time
from functools import singledispatch
from typing import Any, Dict, Optional, Union

import strangeworks as sw
from braket.ahs.analog_hamiltonian_simulation import AnalogHamiltonianSimulation
from braket.annealing.problem import Problem
from braket.circuits import Instruction
from braket.circuits.circuit import Circuit, Gate, QubitSet
from braket.circuits.circuit_helpers import validate_circuit_and_shots
from braket.circuits.compiler_directives import StartVerbatimBox
from braket.circuits.gates import PulseGate
from braket.circuits.serialization import (
    IRType,
    OpenQASMSerializationProperties,
    QubitReferenceType,
    SerializableProgram,
)
from braket.device_schema import GateModelParameters
from braket.device_schema.dwave import (
    Dwave2000QDeviceParameters,
    DwaveAdvantageDeviceParameters,
    DwaveDeviceParameters,
)
from braket.device_schema.dwave.dwave_2000Q_device_level_parameters_v1 import (
    Dwave2000QDeviceLevelParameters,
)
from braket.device_schema.dwave.dwave_advantage_device_level_parameters_v1 import (
    DwaveAdvantageDeviceLevelParameters,
)
from braket.device_schema.ionq import IonqDeviceParameters
from braket.device_schema.oqc import OqcDeviceParameters
from braket.device_schema.rigetti import RigettiDeviceParameters
from braket.device_schema.simulators import GateModelSimulatorDeviceParameters
from braket.error_mitigation import ErrorMitigation
from braket.ir.blackbird import Program as BlackbirdProgram
from braket.ir.openqasm import Program as OpenQASMProgram
from braket.pulse.pulse_sequence import PulseSequence
from braket.schema_common import BraketSchemaBase
from braket.task_result import (
    AnalogHamiltonianSimulationTaskResult,
    AnnealingTaskResult,
    GateModelTaskResult,
    PhotonicModelTaskResult,
)
from braket.tasks import (
    AnalogHamiltonianSimulationQuantumTaskResult,
    AnnealingQuantumTaskResult,
    GateModelQuantumTaskResult,
    PhotonicModelQuantumTaskResult,
    QuantumTask,
)
from strangeworks_core.errors.error import StrangeworksError
from strangeworks_core.types.job import Job, Status


class StrangeworksQuantumTask(QuantumTask):
    _product_slug = "amazon-braket"

    def __init__(self, job: Job, *args, **kwargs):
        self.job: Job = job

    @property
    def id(self) -> str:
        """The id of the task.

        Returns
        -------
        id: str
            The id of the task. This is the id of the job in Strangeworks.
        """
        return self.job.slug

    def cancel(self) -> None:
        """Cancel the task.

        Raises
        ------
        StrangeworksError
            If the task has not been submitted yet.

        """
        if not self.job.external_identifier:
            raise StrangeworksError(
                "Job has not been submitted yet. Missing external_identifier."  # noqa: E501
            )

        resource = sw.get_resource_for_product(StrangeworksQuantumTask._product_slug)
        cancel_url = f"{resource.proxy_url()}/jobs/{self.job.external_identifier}"
        # todo: strangeworks-python is rest_client an optional thing. i dont think it should be # noqa: E501
        # this is something we should discuss
        sw.client.rest_client.delete(url=cancel_url)

    def state(self) -> str:
        """Get the state of the task.

        Returns
        -------
        state: str
            The state of the task.

        Raises
        ------
        StrangeworksError
            If the task has not been submitted yet.
            Or if we find are not able to find the status.
        """
        if not self.job.external_identifier:
            raise StrangeworksError(
                "Job has not been submitted yet. Missing external_identifier."  # noqa: E501
            )

        res = sw.execute_get(
            StrangeworksQuantumTask._product_slug,
            f"jobs/{self.job.external_identifier}",
        )
        self.job = StrangeworksQuantumTask._transform_dict_to_job(res)

        if not self.job.remote_status:
            raise StrangeworksError("Job has no status")
        return self.job.remote_status

    def result(self) -> Union[GateModelQuantumTaskResult, AnnealingQuantumTaskResult]:
        """Get the result of the task.

        Returns
        -------
        result: Union[GateModelQuantumTaskResult, AnnealingQuantumTaskResult]
            The result of the task.

        Raises
        ------
        StrangeworksError
            If the task has not been submitted yet.
            Or if the task did not complete successfully.
            Or unable to fetch the results for the task.
        """
        if not self.job.external_identifier:
            raise StrangeworksError(
                "Job has not been submitted yet. Missing external_identifier."  # noqa: E501
            )
        while self.job.status not in {
            Status.COMPLETED,
            Status.FAILED,
            Status.CANCELLED,
        }:
            res = sw.execute_get(
                StrangeworksQuantumTask._product_slug,
                f"jobs/{self.job.external_identifier}",
            )
            self.job = StrangeworksQuantumTask._transform_dict_to_job(res)
            time.sleep(2.5)

        if self.job.status != Status.COMPLETED:
            raise StrangeworksError("Job did not complete successfully")
        # sw.jobs will return type errors until it updates their type hints
        # todo: update strangeworks-python job type hints
        # todo: at this point in time, sw.jobs returns a different type than sw.execute
        jobs = sw.jobs(slug=self.job.slug)
        if not jobs:
            raise StrangeworksError("Job not found.")
        if len(jobs) != 1:
            raise StrangeworksError("Multiple jobs found.")
        job: Job = jobs[0]
        if not job.files:
            raise StrangeworksError("Job has no files.")
        # for now the strangeworks-python library still returns the Job.files as Files not JobFiles # noqa: E501
        files = list(
            filter(lambda f: f.file_name == "job_results_braket.json", job.files)
        )
        if len(files) != 1:
            raise StrangeworksError("Job has multiple files")

        file = files[0]
        if not file.url:
            raise StrangeworksError("Job file has no url")
        # why does this say it returns a list of files?
        # did it not just download the file?
        # is the contents not some dictionary?
        # todo probably have to update this in strangeworks-python
        contents = sw.download_job_files([file.url])
        if not contents:
            raise StrangeworksError("Unable to download result file.")
        if len(contents) != 1:
            raise StrangeworksError("Unable to download result file.")
        bsh = BraketSchemaBase.parse_raw_schema(json.dumps(contents[0]))

        if (
            bsh.taskMetadata.deviceId
            != "arn:aws:braket:us-east-1::device/qpu/quera/Aquila"
            and bsh.taskMetadata.deviceId
            != "arn:aws:braket:us-east-1::device/qpu/xanadu/Borealis"
        ):
            task_result = GateModelQuantumTaskResult.from_object(bsh)
        else:
            task_result = bsh
        return task_result

    def async_result(self) -> asyncio.Task:
        raise NotImplementedError

    def metadata(self, use_cached_value: bool = False) -> Dict[str, Any]:
        raise NotImplementedError

    @staticmethod
    def from_strangeworks_slug(id: str) -> StrangeworksQuantumTask:
        """Get a task from a strangeworks id.

        Parameters
        ----------
        id: str
            The strangeworks id of the task.

        Returns
        -------
        task: StrangeworksQuantumTask
            The task.

        Raises
        ------
        StrangeworksError
            If no task is found for the id.
            Or if multiple tasks are found for the id.
        """
        # todo: at this point in time, sw.jobs returns a different type than sw.execute
        jobs = sw.jobs(slug=id)
        if not jobs:
            raise StrangeworksError("No jobs found for slug")
        if len(jobs) != 1:
            raise StrangeworksError("Multiple jobs found for slug")
        job = jobs[0]
        return StrangeworksQuantumTask(job)

    @staticmethod
    def create(
        device_arn: str,
        task_specification: Union[
            Circuit,
            Problem,
            OpenQASMProgram,
            BlackbirdProgram,
            PulseSequence,
            AnalogHamiltonianSimulation,
        ],
        shots: int,
        device_parameters: dict[str, Any] | None = None,
        disable_qubit_rewiring: bool = False,
        tags: dict[str, str] | None = None,
        inputs: dict[str, float] | None = None,
        gate_definitions: dict[tuple[Gate, QubitSet], PulseSequence] | None = None,
        quiet: bool = False,
        reservation_arn: str | None = None,
        *args,
        **kwargs,
    ) -> StrangeworksQuantumTask:
        """AwsQuantumTask factory method that serializes a quantum task specification
        (either a quantum circuit or annealing problem), submits it to Amazon Braket,
        and returns back an AwsQuantumTask tracking the execution.

        Args:
            aws_session (AwsSession): AwsSession to connect to AWS with.

            device_arn (str): The ARN of the quantum device.

            task_specification (Union[Circuit, Problem, OpenQASMProgram, BlackbirdProgram, PulseSequence, AnalogHamiltonianSimulation]): # noqa
                The specification of the quantum task to run on device.

            s3_destination_folder (AwsSession.S3DestinationFolder): NamedTuple, with bucket
                for index 0 and key for index 1, that specifies the Amazon S3 bucket and folder
                to store quantum task results in.

            shots (int): The number of times to run the quantum task on the device. If the device is
                a simulator, this implies the state is sampled N times, where N = `shots`.
                `shots=0` is only available on simulators and means that the simulator
                will compute the exact results based on the quantum task specification.

            device_parameters (dict[str, Any] | None): Additional parameters to send to the device.

            disable_qubit_rewiring (bool): Whether to run the circuit with the exact qubits chosen,
                without any rewiring downstream, if this is supported by the device.
                Only applies to digital, gate-based circuits (as opposed to annealing problems).
                If ``True``, no qubit rewiring is allowed; if ``False``, qubit rewiring is allowed.
                Default: False

            tags (dict[str, str] | None): Tags, which are Key-Value pairs to add to this quantum
                task. An example would be:
                `{"state": "washington"}`

            inputs (dict[str, float] | None): Inputs to be passed along with the
                IR. If the IR supports inputs, the inputs will be updated with this value.
                Default: {}.

            gate_definitions (dict[tuple[Gate, QubitSet], PulseSequence] | None): A `dict`
                of user defined gate calibrations. Each calibration is defined for a particular
                `Gate` on a particular `QubitSet` and is represented by a `PulseSequence`.
                Default: None.

            quiet (bool): Sets the verbosity of the logger to low and does not report queue
                position. Default is `False`.

            reservation_arn (str | None): The reservation ARN provided by Braket Direct
                to reserve exclusive usage for the device to run the quantum task on.
                Note: If you are creating tasks in a job that itself was created reservation ARN,
                those tasks do not need to be created with the reservation ARN.
                Default: None.

        Returns:
            AwsQuantumTask: AwsQuantumTask tracking the quantum task execution on the device.

        Note:
            The following arguments are typically defined via clients of Device.
                - `task_specification`
                - `s3_destination_folder`
                - `shots`

        See Also:
            `braket.aws.aws_quantum_simulator.AwsQuantumSimulator.run()`
            `braket.aws.aws_qpu.AwsQpu.run()`
        """  # noqa E501

        create_task_kwargs = _create_common_params(
            device_arn,
            shots if shots is not None else 100,
        )

        if tags is not None:
            create_task_kwargs.update({"tags": tags})
        inputs = inputs or {}
        gate_definitions = gate_definitions or {}

        if reservation_arn:
            create_task_kwargs.update(
                {
                    "associations": [
                        {
                            "arn": reservation_arn,
                            "type": "RESERVATION_TIME_WINDOW_ARN",
                        }
                    ]
                }
            )

        if isinstance(task_specification, Circuit):
            param_names = {param.name for param in task_specification.parameters}
            if unbounded_parameters := param_names - set(inputs.keys()):
                raise ValueError(
                    f"Cannot execute circuit with unbound parameters: {unbounded_parameters}"  # noqa
                )

        create_task_kwargs = _create_internal(
            task_specification,
            create_task_kwargs,
            device_arn,
            device_parameters or {},
            disable_qubit_rewiring,
            inputs,
            gate_definitions=gate_definitions,
            quiet=quiet,
            *args,
            **kwargs,
        )

        res = sw.execute_post(
            StrangeworksQuantumTask._product_slug, create_task_kwargs, endpoint="runjob"
        )
        sw_job = StrangeworksQuantumTask._transform_dict_to_job(res)
        return StrangeworksQuantumTask(sw_job)

    @staticmethod
    def _transform_dict_to_job(d: Dict[str, Any]) -> Job:
        # create a method that transforms the dict into a job
        # first it must convert the json keys from snake_case to camelCase
        # then it must create a job from the dict
        # todo: this is unfortunate. dont like that we need to do this.
        def to_camel_case(snake_str):
            components = snake_str.split("_")
            # We capitalize the first letter of each component except the first one
            # with the 'title' method and join them together.
            return components[0] + "".join(x.title() for x in components[1:])

        remix = {to_camel_case(key): value for key, value in d.items()}
        return Job.from_dict(remix)


@singledispatch
def _create_internal(
    task_specification: Union[Circuit, Problem, BlackbirdProgram],
    create_task_kwargs: dict[str, Any],
    device_arn: str,
    device_parameters: Union[dict, BraketSchemaBase],
    disable_qubit_rewiring: bool,
    inputs: dict[str, float],
    gate_definitions: dict[tuple[Gate, QubitSet], PulseSequence],
    *args,
    **kwargs,
) -> StrangeworksQuantumTask:
    raise TypeError("Invalid task specification type")


@_create_internal.register
def _(
    pulse_sequence: PulseSequence,
    create_task_kwargs: dict[str, Any],
    device_arn: str,
    _device_parameters: Union[
        dict, BraketSchemaBase
    ],  # Not currently used for OpenQasmProgram
    _disable_qubit_rewiring: bool,
    inputs: dict[str, float],
    gate_definitions: dict[tuple[Gate, QubitSet], PulseSequence],
    *args,
    **kwargs,
) -> StrangeworksQuantumTask:
    openqasm_program = OpenQASMProgram(
        source=pulse_sequence.to_ir(),
        inputs=inputs or {},
    )

    create_task_kwargs["action"] = openqasm_program.json()

    return create_task_kwargs


@_create_internal.register
def _(
    openqasm_program: OpenQASMProgram,
    create_task_kwargs: dict[str, Any],
    device_arn: str,
    device_parameters: Union[dict, BraketSchemaBase],
    _disable_qubit_rewiring: bool,
    inputs: dict[str, float],
    gate_definitions: dict[tuple[Gate, QubitSet], PulseSequence],
    *args,
    **kwargs,
) -> StrangeworksQuantumTask:
    if inputs:
        inputs_copy = (
            openqasm_program.inputs.copy()
            if openqasm_program.inputs is not None
            else {}
        )
        inputs_copy.update(inputs)
        openqasm_program = OpenQASMProgram(
            source=openqasm_program.source,
            inputs=inputs_copy,
        )
    create_task_kwargs["action"] = openqasm_program.json()
    if device_parameters:
        final_device_parameters = (
            _circuit_device_params_from_dict(
                device_parameters,
                device_arn,
                GateModelParameters(qubitCount=0),  # qubitCount unused
            )
            if isinstance(device_parameters, dict)
            else device_parameters
        )
        create_task_kwargs["deviceParameters"] = final_device_parameters.json(
            exclude_none=True
        )

    return create_task_kwargs


@_create_internal.register
def _(
    serializable_program: SerializableProgram,
    create_task_kwargs: dict[str, Any],
    device_arn: str,
    device_parameters: Union[dict, BraketSchemaBase],
    _disable_qubit_rewiring: bool,
    inputs: dict[str, float],
    gate_definitions: Optional[dict[tuple[Gate, QubitSet], PulseSequence]],
    *args,
    **kwargs,
) -> StrangeworksQuantumTask:
    openqasm_program = OpenQASMProgram(
        source=serializable_program.to_ir(ir_type=IRType.OPENQASM)
    )
    return _create_internal(
        openqasm_program,
        create_task_kwargs,
        device_arn,
        device_parameters,
        _disable_qubit_rewiring,
        inputs,
        gate_definitions,
        *args,
        **kwargs,
    )


@_create_internal.register
def _(
    blackbird_program: BlackbirdProgram,
    create_task_kwargs: dict[str, any],
    device_arn: str,
    _device_parameters: Union[dict, BraketSchemaBase],
    _disable_qubit_rewiring: bool,
    inputs: dict[str, float],
    gate_definitions: dict[tuple[Gate, QubitSet], PulseSequence],
    *args,
    **kwargs,
) -> StrangeworksQuantumTask:
    create_task_kwargs["action"] = blackbird_program.json()
    return create_task_kwargs


@_create_internal.register
def _(
    circuit: Circuit,
    create_task_kwargs: dict[str, Any],
    device_arn: str,
    device_parameters: Union[dict, BraketSchemaBase],
    disable_qubit_rewiring: bool,
    inputs: dict[str, float],
    gate_definitions: dict[tuple[Gate, QubitSet], PulseSequence],
    *args,
    **kwargs,
) -> StrangeworksQuantumTask:
    validate_circuit_and_shots(circuit, create_task_kwargs["shots"])
    paradigm_parameters = GateModelParameters(
        qubitCount=circuit.qubit_count, disableQubitRewiring=disable_qubit_rewiring
    )
    final_device_parameters = (
        _circuit_device_params_from_dict(
            device_parameters or {}, device_arn, paradigm_parameters
        )
        if isinstance(device_parameters, dict)
        else device_parameters
    )

    qubit_reference_type = QubitReferenceType.VIRTUAL

    if (
        disable_qubit_rewiring
        or Instruction(StartVerbatimBox()) in circuit.instructions
        or gate_definitions
        or any(
            isinstance(instruction.operator, PulseGate)
            for instruction in circuit.instructions
        )
    ):
        qubit_reference_type = QubitReferenceType.PHYSICAL

    serialization_properties = OpenQASMSerializationProperties(
        qubit_reference_type=qubit_reference_type
    )

    openqasm_program = circuit.to_ir(
        ir_type=IRType.OPENQASM,
        serialization_properties=serialization_properties,
        gate_definitions=gate_definitions,
    )

    if inputs:
        inputs_copy = (
            openqasm_program.inputs.copy()
            if openqasm_program.inputs is not None
            else {}
        )
        inputs_copy.update(inputs)
        openqasm_program = OpenQASMProgram(
            source=openqasm_program.source,
            inputs=inputs_copy,
        )

    create_task_kwargs |= {
        "action": openqasm_program.json(),
        "deviceParameters": final_device_parameters.json(exclude_none=True),
    }
    return create_task_kwargs


@_create_internal.register
def _(
    problem: Problem,
    create_task_kwargs: dict[str, Any],
    device_arn: str,
    device_parameters: Union[
        dict,
        DwaveDeviceParameters,
        DwaveAdvantageDeviceParameters,
        Dwave2000QDeviceParameters,
    ],
    _: bool,
    inputs: dict[str, float],
    gate_definitions: Optional[dict[tuple[Gate, QubitSet], PulseSequence]],
    *args,
    **kwargs,
) -> StrangeworksQuantumTask:
    device_params = _create_annealing_device_params(device_parameters, device_arn)
    create_task_kwargs |= {
        "action": problem.to_ir().json(),
        "deviceParameters": device_params.json(exclude_none=True),
    }

    return create_task_kwargs


@_create_internal.register
def _(
    analog_hamiltonian_simulation: AnalogHamiltonianSimulation,
    create_task_kwargs: dict[str, Any],
    device_arn: str,
    device_parameters: dict,
    _: AnalogHamiltonianSimulationTaskResult,
    inputs: dict[str, float],
    gate_definitions: Optional[dict[tuple[Gate, QubitSet], PulseSequence]],
    *args,
    **kwargs,
) -> StrangeworksQuantumTask:
    create_task_kwargs["action"] = analog_hamiltonian_simulation.to_ir().json()
    return create_task_kwargs


def _circuit_device_params_from_dict(
    device_parameters: dict, device_arn: str, paradigm_parameters: GateModelParameters
) -> GateModelSimulatorDeviceParameters:
    if "errorMitigation" in device_parameters:
        error_migitation = device_parameters["errorMitigation"]
        device_parameters["errorMitigation"] = (
            error_migitation.serialize()
            if isinstance(error_migitation, ErrorMitigation)
            else error_migitation
        )
    if "ionq" in device_arn:
        return IonqDeviceParameters(
            paradigmParameters=paradigm_parameters, **device_parameters
        )
    if "rigetti" in device_arn:
        return RigettiDeviceParameters(paradigmParameters=paradigm_parameters)
    if "oqc" in device_arn:
        return OqcDeviceParameters(paradigmParameters=paradigm_parameters)
    return GateModelSimulatorDeviceParameters(paradigmParameters=paradigm_parameters)


def _create_annealing_device_params(
    device_params: dict[str, Any], device_arn: str
) -> Union[DwaveAdvantageDeviceParameters, Dwave2000QDeviceParameters]:
    """Gets Annealing Device Parameters.

    Args:
        device_params (dict[str, Any]): Additional parameters for the device.
        device_arn (str): The ARN of the quantum device.

    Returns:
        Union[DwaveAdvantageDeviceParameters, Dwave2000QDeviceParameters]: The device parameters.  # noqa

    """
    if not isinstance(device_params, dict):
        device_params = device_params.dict()
    # check for device level or provider level parameters
    device_level_parameters = device_params.get(
        "deviceLevelParameters", None
    ) or device_params.get("providerLevelParameters", {})
    # deleting since it may be the old version
    if "braketSchemaHeader" in device_level_parameters:
        del device_level_parameters["braketSchemaHeader"]

    if "Advantage" in device_arn:
        device_level_parameters = DwaveAdvantageDeviceLevelParameters.parse_obj(
            device_level_parameters
        )
        return DwaveAdvantageDeviceParameters(
            deviceLevelParameters=device_level_parameters
        )
    elif "2000Q" in device_arn:
        device_level_parameters = Dwave2000QDeviceLevelParameters.parse_obj(
            device_level_parameters
        )
        return Dwave2000QDeviceParameters(deviceLevelParameters=device_level_parameters)
    else:
        raise Exception(
            f"Amazon Braket could not find a device with ARN: {device_arn}. "
            "To continue, make sure that the value of the device_arn parameter "
            "corresponds to a valid QPU."
        )


def _create_common_params(device_arn: str, shots: int) -> dict[str, Any]:
    return {
        "deviceArn": device_arn,
        # "outputS3Bucket": s3_destination_folder[0],
        # "outputS3KeyPrefix": s3_destination_folder[1],
        "shots": shots,
    }


@singledispatch
def _format_result(
    result: Union[GateModelTaskResult, AnnealingTaskResult, PhotonicModelTaskResult],
) -> Union[
    GateModelQuantumTaskResult,
    AnnealingQuantumTaskResult,
    PhotonicModelQuantumTaskResult,
]:
    raise TypeError("Invalid result specification type")


@_format_result.register
def _(result: GateModelTaskResult) -> GateModelQuantumTaskResult:
    GateModelQuantumTaskResult.cast_result_types(result)
    return GateModelQuantumTaskResult.from_object(result)


@_format_result.register
def _(result: AnnealingTaskResult) -> AnnealingQuantumTaskResult:
    return AnnealingQuantumTaskResult.from_object(result)


@_format_result.register
def _(result: PhotonicModelTaskResult) -> PhotonicModelQuantumTaskResult:
    return PhotonicModelQuantumTaskResult.from_object(result)


@_format_result.register
def _(
    result: AnalogHamiltonianSimulationTaskResult,
) -> AnalogHamiltonianSimulationQuantumTaskResult:
    return AnalogHamiltonianSimulationQuantumTaskResult.from_object(result)



from __future__ import annotations

import importlib
import json
import urllib.request
from enum import Enum
from typing import Any, Dict, Optional, Union

import strangeworks as sw
from braket.ahs.analog_hamiltonian_simulation import AnalogHamiltonianSimulation
from braket.annealing.problem import Problem
from braket.aws.aws_quantum_task import AwsQuantumTask
from braket.circuits import Circuit, Gate, QubitSet
from braket.circuits.gate_calibrations import GateCalibrations
from braket.circuits.noise_model import NoiseModel
from braket.device_schema import DeviceCapabilities, GateModelQpuParadigmProperties
from braket.device_schema.dwave import DwaveProviderProperties
from braket.device_schema.pulse.pulse_device_action_properties_v1 import (
    PulseDeviceActionProperties,
)
from braket.devices.device import Device
from braket.ir.blackbird import Program as BlackbirdProgram
from braket.ir.openqasm import Program as OpenQasmProgram
from braket.parametric.free_parameter import FreeParameter
from braket.parametric.free_parameter_expression import _is_float
from braket.pulse import ArbitraryWaveform, Frame, Port, PulseSequence
from braket.pulse.waveforms import _parse_waveform_from_calibration_schema
from braket.schema_common import BraketSchemaBase
from braket.tasks.quantum_task import QuantumTask
from braket.tasks.quantum_task_batch import QuantumTaskBatch
from networkx import DiGraph, complete_graph, from_edgelist
from strangeworks_core.errors.error import StrangeworksError

from strangeworks_braket.job import StrangeworksQuantumJob
from strangeworks_braket.task import StrangeworksQuantumTask


class AwsDeviceType(str, Enum):
    """Possible AWS device types"""

    SIMULATOR = "SIMULATOR"
    QPU = "QPU"


class StrangeworksDevice(Device):
    def __init__(
        self,
        arn: str,
        name: Optional[str] = None,
        status: Optional[str] = None,
        slug: Optional[str] = None,
        noise_model: Optional[NoiseModel] = None,
        **kwargs,
    ):
        super().__init__(name, status)
        self._arn = arn
        self.slug = slug
        self._status = status
        self._arn = arn
        self._gate_calibrations = None
        self._properties = None
        self._provider_name = None
        self._poll_interval_seconds = None
        self._type = None
        self._ports = None
        self._frames = None
        if noise_model:
            self._validate_device_noise_model_support(noise_model)
        self._noise_model = noise_model

    def run(
        self,
        task_specification: Union[
            Circuit,
            Problem,
            OpenQasmProgram,
            BlackbirdProgram,
            PulseSequence,
            AnalogHamiltonianSimulation,
        ],
        shots: Optional[int] = None,
        poll_timeout_seconds: float = AwsQuantumTask.DEFAULT_RESULTS_POLL_TIMEOUT,
        poll_interval_seconds: Optional[float] = None,
        inputs: Optional[dict[str, float]] = None,
        gate_definitions: Optional[dict[tuple[Gate, QubitSet], PulseSequence]] = None,
        reservation_arn: str | None = None,
        tags: list[str] = None,
        *aws_quantum_task_args: Any,
        **aws_quantum_task_kwargs: Any,
    ) -> QuantumTask:
        """Run a task on the device.
        Parameters
        ----------
        task_specification: Union[Circuit, Problem, OpenQasmProgram]
            The task specification.
        shots: Optional[int]
            The number of shots to run the task for. Defaults to 1000.
        Returns
        -------
        task: QuantumTask (StrangeworksQuantumTask)
            The task that was run.
        """

        if self._noise_model:
            task_specification = self._apply_noise_model_to_circuit(task_specification)
        return StrangeworksQuantumTask.create(
            self._arn,
            task_specification,
            shots if shots is not None else self._default_shots,
            poll_timeout_seconds=poll_timeout_seconds,
            poll_interval_seconds=poll_interval_seconds or self._poll_interval_seconds,
            inputs=inputs,
            gate_definitions=gate_definitions,
            reservation_arn=reservation_arn,
            tags=tags,
            *aws_quantum_task_args,
            **aws_quantum_task_kwargs,
        )

    def run_hybrid(
        self,
        filepath: str,
        hyperparameters: Dict[str, Any],
        *args,
        **kwargs,
    ) -> QuantumTask:
        """Run a task on the device.
        Parameters
        ----------
        filepath: str
            Path to the python file that will be run.
        hyperparameters: Dict[str, Any]
            Dictionary of hyperparameters to pass to the task.
            Must be json serializable.
        Returns
        -------
        Job: QuantumJob (StrangeworksQuantumJob)
            The job that was run.
        """
        return StrangeworksQuantumJob.create_hybrid(
            self._arn, filepath, hyperparameters, *args, **kwargs
        )

    @property
    def status(self) -> str:
        if self._status is None:
            dev = self.get_devices(arns=[self._arn])[0]
            self._status = dev.status

        return self._status

    @staticmethod
    def get_devices(
        arns: Optional[list[str]] = None,
        names: Optional[list[str]] = None,
        statuses: Optional[list[str]] = None,
    ) -> list[StrangeworksDevice]:
        """Get a list of devices.
        Parameters
        ----------
        arns: Optional[list[str]
            Filter by list of device ARNs. Defaults to None.
        names: Optional[list[str]]
            Filter by list of device names. Defaults to None.
        statuses: Optional[list[str]]
            Filter by list of device statuses. Defaults to None.
        Returns
        -------
        devices: list[SwDevice]
            List of devices that match the provided filters.
        """
        backends = sw.backends(product_slugs=["amazon-braket"])
        devices = []
        for backend in backends:
            if arns and backend.remote_backend_id not in arns:
                continue
            if names and backend.name not in names:
                continue
            if statuses and backend.remote_status not in statuses:
                continue

            devices.append(
                StrangeworksDevice(
                    backend.remote_backend_id,
                    backend.name,
                    backend.remote_status,
                    backend.slug,
                )
            )

        return devices

    def run_batch(
        self,
        task_specifications: Circuit | Problem | list[Circuit | Problem],
        shots: int | None,
        max_parallel: int | None,
        inputs: Dict[str, float] | list[Dict[str, float]] | None,
        *args,
        **kwargs,
    ) -> QuantumTaskBatch:
        raise StrangeworksError("currently not implemented/supported")

    def refresh_metadata(self) -> None:
        """Refresh the `AwsDevice` object with the most recent Device metadata."""
        self._populate_properties()

    def _populate_properties(self) -> None:
        payload = {
            "aws_device_arn": self._arn,
        }
        metadata = sw.execute_post(
            StrangeworksQuantumTask._product_slug, payload, endpoint="metadata"
        )
        self._name = metadata.get("deviceName")
        self._status = metadata.get("deviceStatus")
        self._type = AwsDeviceType(metadata.get("deviceType"))
        self._provider_name = metadata.get("providerName")
        self._properties = BraketSchemaBase.parse_raw_schema(
            metadata.get("deviceCapabilities")
        )
        device_poll_interval = self._properties.service.getTaskPollIntervalMillis
        self._poll_interval_seconds = (
            device_poll_interval / 1000.0
            if device_poll_interval
            else AwsQuantumTask.DEFAULT_RESULTS_POLL_INTERVAL
        )
        self._topology_graph = None
        self._frames = None
        self._ports = None

    @property
    def provider_name(self) -> str:
        """str: Return the provider name"""
        return self._provider_name

    @property
    def arn(self) -> str:
        """str: Return the ARN of the device"""
        return self._arn

    @property
    def gate_calibrations(self) -> Optional[GateCalibrations]:
        """Calibration data for a QPU. Calibration data is shown for gates on particular gubits.  # noqa
        If a QPU does not expose these calibrations, None is returned.

        Returns:
            Optional[GateCalibrations]: The calibration object. Returns `None` if the data  # noqa
            is not present.
        """
        ###########################
        # ToDo:
        # Create Gate calibration call
        ###########################
        if not self._gate_calibrations:
            self._gate_calibrations = self.refresh_gate_calibrations()
        return self._gate_calibrations

    @property
    def properties(self) -> DeviceCapabilities:
        """DeviceCapabilities: Return the device properties

        Please see `braket.device_schema` in amazon-braket-schemas-python_

        .. _amazon-braket-schemas-python: https://github.com/aws/amazon-braket-schemas-python  # noqa
        """
        if self._properties is None:
            self._populate_properties()

        return self._properties

    @property
    def topology_graph(self) -> DiGraph:
        """DiGraph: topology of device as a networkx `DiGraph` object.

        Examples:
            >>> import networkx as nx
            >>> device = AwsDevice("arn1")
            >>> nx.draw_kamada_kawai(device.topology_graph, with_labels=True, font_weight="bold")  # noqa

            >>> topology_subgraph = device.topology_graph.subgraph(range(8))
            >>> nx.draw_kamada_kawai(topology_subgraph, with_labels=True, font_weight="bold")  # noqa

            >>> print(device.topology_graph.edges)

        Returns:
            DiGraph: topology of QPU as a networkx `DiGraph` object. `None` if the topology  # noqa
            is not available for the device.
        """
        if not self._topology_graph:
            self._topology_graph = self._construct_topology_graph()
        return self._topology_graph

    def _construct_topology_graph(self) -> DiGraph:
        """Construct topology graph. If no such metadata is available, return `None`.

        Returns:
            DiGraph: topology of QPU as a networkx `DiGraph` object.
        """
        if hasattr(self.properties, "paradigm") and isinstance(
            self.properties.paradigm, GateModelQpuParadigmProperties
        ):
            if self.properties.paradigm.connectivity.fullyConnected:
                return complete_graph(
                    int(self.properties.paradigm.qubitCount), create_using=DiGraph()
                )
            adjacency_lists = self.properties.paradigm.connectivity.connectivityGraph
            edges = []
            for item in adjacency_lists.items():
                i = item[0]
                edges.extend([(int(i), int(j)) for j in item[1]])
            return from_edgelist(edges, create_using=DiGraph())
        elif hasattr(self.properties, "provider") and isinstance(
            self.properties.provider, DwaveProviderProperties
        ):
            edges = self.properties.provider.couplers
            return from_edgelist(edges, create_using=DiGraph())
        else:
            return None

    @property
    def frames(self) -> dict[str, Frame]:
        """Returns a dict mapping frame ids to the frame objects for predefined frames
        for this device.
        """
        self._update_pulse_properties()
        return self._frames or {}

    @property
    def ports(self) -> dict[str, Port]:
        """Returns a dict mapping port ids to the port objects for predefined ports
        for this device.
        """
        self._update_pulse_properties()
        return self._ports or {}

    def _update_pulse_properties(self) -> None:
        if not hasattr(self.properties, "pulse") or not isinstance(
            self.properties.pulse, PulseDeviceActionProperties
        ):
            return
        if self._ports is None:
            self._ports = {}
            port_data = self.properties.pulse.ports
            for port_id, port in port_data.items():
                self._ports[port_id] = Port(
                    port_id=port_id, dt=port.dt, properties=json.loads(port.json())
                )
        if self._frames is None:
            self._frames = {}
            if frame_data := self.properties.pulse.frames:
                for frame_id, frame in frame_data.items():
                    self._frames[frame_id] = Frame(
                        frame_id=frame_id,
                        port=self._ports[frame.portId],
                        frequency=frame.frequency,
                        phase=frame.phase,
                        is_predefined=True,
                        properties=json.loads(frame.json()),
                    )

    def refresh_gate_calibrations(self) -> Optional[GateCalibrations]:
        """Refreshes the gate calibration data upon request.

        If the device does not have calibration data, None is returned.

        Raises:
            URLError: If the URL provided returns a non 2xx response.

        Returns:
            Optional[GateCalibrations]: the calibration data for the device. None
            is returned if the device does not have a gate calibrations URL associated.
        """
        ###########################
        # ToDo:
        # Create Gate calibration call
        ###########################
        if (
            hasattr(self.properties, "pulse")
            and hasattr(self.properties.pulse, "nativeGateCalibrationsRef")
            and self.properties.pulse.nativeGateCalibrationsRef
        ):
            try:
                with urllib.request.urlopen(
                    self.properties.pulse.nativeGateCalibrationsRef.split("?")[0]
                ) as f:
                    json_calibration_data = self._parse_calibration_json(
                        json.loads(f.read().decode("utf-8"))
                    )
                    return GateCalibrations(json_calibration_data)
            except urllib.error.URLError as e:
                raise urllib.error.URLError(
                    f"Unable to reach {self.properties.pulse.nativeGateCalibrationsRef}"
                ) from e
        else:
            return None

    def _parse_waveforms(self, waveforms_json: dict) -> dict:
        waveforms = {}
        for waveform in waveforms_json:
            parsed_waveform = _parse_waveform_from_calibration_schema(
                waveforms_json[waveform]
            )
            waveforms[parsed_waveform.id] = parsed_waveform
        return waveforms

    def _parse_pulse_sequence(
        self, calibration: dict, waveforms: dict[ArbitraryWaveform]
    ) -> PulseSequence:
        return PulseSequence._parse_from_calibration_schema(
            calibration, waveforms, self.frames
        )

    def _parse_calibration_json(
        self, calibration_data: dict
    ) -> dict[tuple[Gate, QubitSet], PulseSequence]:
        """Takes the json string from the device calibration URL and returns a structured dictionary of
        corresponding `dict[tuple[Gate, QubitSet], PulseSequence]` to represent the calibration data.

        Args:
            calibration_data (dict): The data to be parsed. Based on
                https://github.com/aws/amazon-braket-schemas-python/blob/main/src/braket/device_schema/pulse/native_gate_calibrations_v1.py.

        Returns:
            dict[tuple[Gate, QubitSet], PulseSequence]: The
            structured data based on a mapping of `tuple[Gate, Qubit]` to its calibration represented as a
            `PulseSequence`.

        """  # noqa: E501
        waveforms = self._parse_waveforms(calibration_data["waveforms"])
        parsed_calibration_data = {}
        for qubit_node in calibration_data["gates"]:
            qubit = calibration_data["gates"][qubit_node]
            for gate_node in qubit:
                for gate in qubit[gate_node]:
                    gate_capitalized = getattr(
                        self,
                        f"_{self.provider_name.upper()}_GATES_TO_BRAKET",
                        {},
                    ).get(gate_node.capitalize(), gate_node.capitalize())

                    if gate_capitalized == "Iswap":
                        gate_capitalized = "ISwap"
                    elif gate_capitalized == "Cz":
                        gate_capitalized = "CZ"
                    elif gate_capitalized == "Gphase":
                        gate_capitalized = "GPhase"
                    elif gate_capitalized == "Phaseshift":
                        gate_capitalized = "PhaseShift"
                    elif gate_capitalized == "Cnot":
                        gate_capitalized = "CNot"
                    elif gate_capitalized == "Pswap":
                        gate_capitalized = "PSwap"
                    elif gate_capitalized == "Xy":
                        gate_capitalized = "XY"
                    elif gate_capitalized == "Cphaseshift":
                        gate_capitalized = "CPhaseShift"
                    elif gate_capitalized == "Cphaseshift00":
                        gate_capitalized = "CPhaseShift00"
                    elif gate_capitalized == "Cphaseshift01":
                        gate_capitalized = "CPhaseShift01"
                    elif gate_capitalized == "Cphaseshift10":
                        gate_capitalized = "CPhaseShift10"
                    elif gate_capitalized == "Cv":
                        gate_capitalized = "CV"
                    elif gate_capitalized == "Cy":
                        gate_capitalized = "CY"
                    elif gate_capitalized == "Ecr":
                        gate_capitalized = "ECR"
                    elif gate_capitalized == "Xx":
                        gate_capitalized = "XX"
                    elif gate_capitalized == "Yy":
                        gate_capitalized = "YY"
                    elif gate_capitalized == "Zz":
                        gate_capitalized = "ZZ"
                    elif gate_capitalized == "Ccnot":
                        gate_capitalized = "CCNot"
                    elif gate_capitalized == "Cswap":
                        gate_capitalized = "CSwap"
                    elif gate_capitalized == "Gpi":
                        gate_capitalized = "GPi"
                    elif gate_capitalized == "Prx":
                        gate_capitalized = "PRx"
                    elif gate_capitalized == "Gpi2":
                        gate_capitalized = "GPi2"
                    elif gate_capitalized == "Ms":
                        gate_capitalized = "MS"
                    elif gate_capitalized == "Pulsegate":
                        gate_capitalized = "PulseGate"

                    gate_obj = (
                        getattr(
                            importlib.import_module("braket.circuits.gates"),
                            gate_capitalized,
                        )
                        if gate_capitalized is not None
                        else None
                    )
                    qubits = QubitSet([int(x) for x in gate["qubits"]])
                    if gate_obj is None:
                        # We drop out gates that are not implemented in the BDK
                        continue

                    argument = None
                    if gate["arguments"]:
                        argument = (
                            float(gate["arguments"][0])
                            if _is_float(gate["arguments"][0])
                            else FreeParameter(gate["arguments"][0])
                        )
                    gate_qubit_key = (
                        (gate_obj(argument), qubits)
                        if argument
                        else (gate_obj(), qubits)
                    )
                    gate_qubit_pulse = self._parse_pulse_sequence(
                        gate["calibrations"], waveforms
                    )
                    parsed_calibration_data[gate_qubit_key] = gate_qubit_pulse

        return parsed_calibration_data



from typing import List, Tuple

import numpy as np
from braket.timings.time_series import TimeSeries


def rabi_pulse(
    rabi_pulse_area: float, omega_max: float, omega_slew_rate_max: float
) -> Tuple[List[float], List[float]]:
    """Get a time series for Rabi frequency with specified Rabi phase, maximum amplitude
    and maximum slew rate

        Args:
            rabi_pulse_area (float): Total area under the Rabi frequency time series
            omega_max (float): The maximum amplitude
            omega_slew_rate_max (float): The maximum slew rate

        Returns:
            Tuple[List[float], List[float]]: A tuple containing the time points and
            values
                of the time series for the time dependent Rabi frequency

        Notes: By Rabi phase, it means the integral of the amplitude of a time-dependent
            Rabi frequency, int_0^TOmega(t)dt, where T is the duration.
    """

    phase_threshold = omega_max**2 / omega_slew_rate_max
    if rabi_pulse_area <= phase_threshold:
        t_ramp = np.sqrt(rabi_pulse_area / omega_slew_rate_max)
        t_plateau = 0
    else:
        t_ramp = omega_max / omega_slew_rate_max
        t_plateau = (rabi_pulse_area / omega_max) - t_ramp
    t_pules = 2 * t_ramp + t_plateau
    time_points = [0, t_ramp, t_ramp + t_plateau, t_pules]
    amplitude_values = [
        0,
        t_ramp * omega_slew_rate_max,
        t_ramp * omega_slew_rate_max,
        0,
    ]

    return time_points, amplitude_values


def constant_time_series(
    other_time_series: TimeSeries, constant: float = 0.0
) -> TimeSeries:
    """Obtain a constant time series with the same time points as the given time series

    Args:
        other_time_series (TimeSeries): The given time series

    Returns:
        TimeSeries: A constant time series with the same time points as the given time
        series
    """
    ts = TimeSeries()
    for t in other_time_series.times():
        ts.put(t, constant)
    return ts
